// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: type_impl.cc,v 1.2 2008-08-21 15:34:10 carzanig Exp $
//
#include <cassert>
#include <iostream>
#include <vector>
#include "siena/simplify.h"
#include "siena/ssimp_types.h"

using namespace siena;
using namespace std;

int main( int argc, char** argv )
{
  // used for covering
  Constraint c1( string( "x" ), eq_id, (int_t)10 );
  Constraint c1a( c1 );
  Constraint c2( string( "x" ), lt_id, (int_t)11 );

  // used for conflicting
  Constraint x1( string( "x" ), eq_id, (int_t)10 );
  Constraint x2( string( "x" ), lt_id, (int_t)5 );
  Constraint x3( string( "x" ), lt_id, (int_t)20 );
  Constraint x4( string( "x" ), gt_id, (int_t)12 );
  Constraint x5( string( "x" ), gt_id, (int_t)5 );
  Constraint x6( string( "x" ), eq_id, (siena::double_t)3.141 );


  
  // setup an initial filter with c1 and add c2 which is covered by c1
  SimplifyingFilter f1;
  f1.add( c1 );
  assert( f1.size() == 1 );
  assert( f1.bytesize() == c1.bytesize() );
  f1.add( c1a );
  assert( f1.size() == 1 );
  assert( f1.bytesize() == c1.bytesize() );
  
  f1.add( c2 );
  assert( f1.size() == 1 );
  assert( f1.getConstraints().find( c2 ) != f1.getConstraints().end() );
  assert( f1.bytesize() == c2.bytesize() );

  // and add in reverse order to check opposite covering relation
  SimplifyingFilter f2;
  f2.add( c2 );
  assert( f2.size() == 1 );
  assert( f2.bytesize() == c2.bytesize() );
  
  f2.add( c1 );
  assert( f2.size() == 1 );
  assert( f2.getConstraints().find( c2 ) != f2.getConstraints().end() );
  assert( f2.bytesize() == c2.bytesize() );

  // test conflicting
  SimplifyingFilter f3;
  f3.add( x1 );
  assert( f3.size() == 1 );
  assert( f3.bytesize() == x1.bytesize() );
  assert( conflicting( x1, x2 ) );
  f3.add( x2 ); // conflicts
  assert( f3.size() == 0 );
  assert( f3.bytesize() == 0 );
  f3.add( x1 ); // won't add any more now
  assert( f3.size() == 0 );
  assert( f3.bytesize() == 0 );

  // test conflicting in opposite order
  SimplifyingFilter f4;
  f4.add( x2 );
  assert( f4.size() == 1 );
  assert( f4.bytesize() == x2.bytesize() );
  assert( conflicting( x1, x2 ) );
  f4.add( x1 ); // conflicts
  assert( f4.size() == 0 );
  assert( f4.bytesize() == 0 );
  f4.add( x2 ); // won't add any more now
  assert( f4.size() == 0 );
  assert( f4.bytesize() == 0 );

  // now test predicates
  SimplifyingPredicate p1;

  SimplifyingFilter f5;

  p1.add( f5 ); // shouldn't add an empty filter
  assert( p1.size() == 0 );
  assert( p1.bytesize() == 0 );

  f5.add( c1 );

  p1.add( f5 );
  assert( p1.size() == 1 );
  assert( p1.bytesize() == f5.bytesize() );

  p1.add( f5 );
  assert( p1.size() == 1 );
  assert( p1.bytesize() == f5.bytesize() );

  Constraint y1( string( "y" ), eq_id, (int_t)10 );

  SimplifyingFilter f6;
  f6.add( c1 );
  f6.add( y1 );
  assert( f6.size() == 2 );
  assert( f6.bytesize() == (c1.bytesize() + y1.bytesize()) );

  // filter f5 covers f6
  p1.add( f6 );
  assert( p1.size() == 1 );
  assert( p1.getFilters().find( f5 ) != p1.getFilters().end() );
  assert( p1.bytesize() == f5.bytesize() );

  // add a unrelated constraint
  SimplifyingFilter f7;
  f7.add( y1 );

  p1.add( f7 );
  assert( p1.size() == 2 );
  assert( p1.bytesize() == f5.bytesize() + f7.bytesize() );


  /*
   * test some other operators
   */
  SimplifyingPredicate pA;
  pA.add( f6 );
  pA.add( f7 );

  SimplifyingPredicate pB;

  pB = pA;
  assert( pA == pB );
  assert( pB == pA );

  SimplifyingPredicate pC;

  pA = pC;

  assert( pA == pC );
  assert( pC == pA );
}
